\\ Copyright 2023 Kevin Ryde

\\ This file is part of pari-primecount.

\\ pari-primecount is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ pari-primecount is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with pari-primecount.  If not, see <http://www.gnu.org/licenses/>.


\\ Usage: read("primecount.gp");
\\
\\ This is a GP interface to the primecount library or program.
\\ Primecount has primepi() and prime() equivalents which may
\\ allow larger calculations than native PARI.
\\
\\ Functions are named like the primecount library,
\\
\\        n = primecount_pi(x)           like primepi()
\\        x = primecount_get_max_x()     pi limit (for the library)
\\        p = primecount_nth_prime(n)    like prime()
\\        n = primecount_phi(x,a)
\\
\\        str = primecount_version()
\\        n = primecount_get_num_threads()
\\        primecount_set_num_threads(n)
\\
\\ To install, compile pari_primecount.c to shared library
\\ pari_primecount.so (see Makefile) and put it somewhere GP
\\ install() can find it, which usually means
\\
\\     default(sopath)   directory list
\\     LD_LIBRARY_PATH   environment variable directory list
\\     /usr/lib          or similar usual directory
\\
\\ To use the primecount program, have that program somewhere
\\ in the usual execution PATH ready to run.
\\
\\ Running the primecount program each time is slower than
\\ library calls, so the library is preferred.  To force use of
\\ the program, set GP variable primecount_use_program before
\\ loading
\\
\\     primecount_use_program = 1;
\\     read("primecount.gp");
\\
\\ The program has some options which are not in the library
\\ interface, but think the library forms are best for general use.
\\
\\
\\ Bugs
\\ ----
\\
\\ Auto-detection of interface pari_primecount.so not available
\\ (so use the program instead) prints a spurious warning message
\\ to stderr.
\\
\\ The library interface throws PARI e_DOMAIN or e_OVERFLOW
\\ errors as appropriate, but currently the program interface
\\ prints to stderr and just GP user level error().
\\
\\ In the library interface, large x in primecount_phi() might
\\ cause the library to fail to malloc enough memory and which
\\ results in its error exit.  Would prefer PARI e_MEM (assuming
\\ recovery is in fact possible).
\\
\\
\\ Changes
\\ -------
\\
\\ Version 1 - the first version.


if(primecount_INTERNAL_verbose=='primecount_INTERNAL_verbose, \
   primecount_INTERNAL_verbose=0);

{
  if(pari_primecount_version != 'pari_primecount_version,
     \\ already loaded,
     \\ avoid install() errors on attempting to repeat load
     ,

     (primecount_use_program=='primecount_use_program
       || ! primecount_use_program)
     && iferr(install("pari_primecount_init","v",
                      "pari_primecount_init",
                      "pari_primecount.so"); 1,
              e, 0),
     if(primecount_INTERNAL_verbose,
        print("primecount using pari_primecount.so"));

     \\ eval() so the parser doesn't see pari_primecount_init
     \\ symbol before executing the install() above
     eval("pari_primecount_init()");

     install("pari_primecount_version","",
             "primecount_version",
             "pari_primecount.so");

     install("pari_primecount_get_num_threads","",
             "primecount_get_num_threads",
             "pari_primecount.so");
     install("pari_primecount_set_num_threads","vG",
             "primecount_set_num_threads",
             "pari_primecount.so");

     install("pari_primecount_pi","G",
             "primecount_pi","pari_primecount.so");
     install("pari_primecount_get_max_x","",
             "primecount_get_max_x","pari_primecount.so");

     install("pari_primecount_nth_prime","G",
             "primecount_nth_prime","pari_primecount.so");
     install("pari_primecount_phi","GG",
             "primecount_phi","pari_primecount.so");

     ,
\\-----------
     if(primecount_INTERNAL_verbose,
        print("primecount using primecount program"));

my(digit_set = Set(Vecsmall("0123456789")));
(
\\ v is a vector of strings from externstr()
\\ demand a single integer number (and so safe to eval())
primecount_INTERNAL_validate(v) =
  my(good = (#v==1));
  if(good,
     v=Vecsmall(v[1]);
     for(i=1,#v,
        if(! setsearch(digit_set,v[i]),
           good=0; break)));
  good || error("unrecognised output from primecount program");
);

primecount_INTERNAL_num_threads = 0;
(
primecount_version() =
  my(v=externstr("primecount --version"));
  if(#v, concat(apply(s->Str(s,"\n"), v)), "");
);
(
primecount_INTERNAL_program_run(operation,n) =
  \\ --status might have been useful for verbose on slow operation,
  \\ but its output goes to stdout the same as the normal result,
  \\ so difficult to separate

  my(str=Str("primecount",
             if(primecount_INTERNAL_num_threads,
                Str(" --threads="primecount_INTERNAL_num_threads),
                ""),
             " ",operation," ",n));
  if(primecount_INTERNAL_verbose, print(str));
  my(v=externstr(str));
  primecount_INTERNAL_validate(v);
  eval(v[1]);
);
(
primecount_nth_prime(n) = primecount_INTERNAL_program_run("-n",n);
);
(
primecount_pi(n) =
  n=floor(n);
  if(n<=0,0,  primecount_INTERNAL_program_run("",n));
);
(
primecount_phi(x,a) =
  x=floor(x);
  if(x<=0||a<=0,0,
     primecount_INTERNAL_program_run("--phi",Str(x" "a)));
);
(
primecount_get_num_threads() =
  primecount_INTERNAL_num_threads;
);
(
primecount_set_num_threads(t) =
  type(t)=="t_INT" || error();
  primecount_INTERNAL_num_threads = t;
);
);
}

pari_primecount_version = 1;

{
addhelp(primecount_version,
"str = primecount_version()
Return a string with the primecount library version number,
or program version string when using the primecount program.");
}

{
addhelp(primecount_get_num_threads,
"n = primecount_get_num_threads()
Return the number of threads which primecount should use.
The default is the number of CPU cores.
With the primecount program, the default (all cores) is returned
as 0, rather than the actual number.");
}
{
addhelp(primecount_set_num_threads,
"primecount_set_num_threads(n)
Set the number of threads which primecount should use.
The default is the number of CPU cores.");
}

{
addhelp(primecount_pi,
"n = primecount_pi(x)
Return the number of primes <= x, like PARI primepi(x).
x can be a fraction.
Must have x < 2^63 on a 32-bit system or x < 10^31 on a 64-bit system.
See primecount_get_max_x() for this maximum.");
}

\\ no function primecount_get_max_x() when using the program interface
\\ MY-CHECK-GP-FUNCTION-DEFINITION: primecount_get_max_x()
{
addhelp(primecount_get_max_x,
"n = primecount_get_max_x()
Return the largest x allowed in primecount_pi().
This function is only available with the primecount library,
not with the primecount program.
At the time of writing, max x is 2^63 - 1 on a 32-bit system,
or 10^31 - 1 on a 64-bit system.");
}

{
addhelp(primecount_nth_prime,
"prime = primecount_nth_prime(n)
Return the n'th prime, like PARI prime(n).
The first prime is n=1 prime 2.
n is limited to the resulting prime < 2^63,
which means n <= 216289611853439384.
.");
}

{
addhelp(primecount_phi,
"n = primecount_phi(x,a)
Return the number of integers 1..x which are not divisible
by any of the first \"a\" primes.
x can be a fraction.  Must have x < 2^63.");
}
